/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <string.h>

#include "OO7.hxx"
#include "GenParams.h"
#include "VarParams.h"
#include "baidlist.h"

extern int	nextAtomicId;
extern int	debugMode;
extern char*	types[NumTypes];

extern BAIdList* private_cp;
extern BAIdList* shared_cp;

extern Ref<Cluster> DocCluster ;

////////////////////////////////////////////////////////////////////////////
//
// CompositePart Constructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::CompositePart(int cpId) : id(cpId)
{
   BaseAssembly* ba;

   if (debugMode) {
        printf("BEGIN CompositePart::CompositePart(cpId = %d)\n", cpId);
   }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  init data members\n");
   }

    // initialize this composite part's simple data members (some randomly)

    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);

    // for the build date, decide if this part is young or old, and then
    // randomly choose a date in the required range

    if (cpId % YoungCompFrac == 0) {
        // young one
        if (debugMode) {
            fprintf(stderr, "(young composite part, id = %d.)\n", id);
        }
        buildDate = MinYoungCompDate + 
		    (int)(random() % (MaxYoungCompDate - MinYoungCompDate + 1));
    } else { 
        // old one
        if (debugMode) {
            fprintf(stderr, "(old composite part, id = %d.)\n", id);
        }
        buildDate = MinOldCompDate + 
		    (int)(random() % (MaxOldCompDate - MinOldCompDate + 1));
    }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  update extent\n");
   }

    // add part to CompositePart extent

    CompositePart::extent->insert(this) ;

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  create document\n");
   }

    // initialize the documentation (indexed by its title) ...

   documentation = new(DocCluster) Document (cpId);
   documentation->part = this ;

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  create atomic parts\n");
   }

    // now create the atomic parts (indexed by their ids) ...

   int i, id, from, to;
   AtomicPart** atom = new AtomicPart* [NumAtomicPerComp];
   Connection* conn;

   this->parts.init(NumAtomicPerComp) ;
    for (i = 0; i < NumAtomicPerComp; i++) {
	id = nextAtomicId + i;
	atom[i] = new (this) AtomicPart (id);
	parts += atom[i];
	atom[i]->partOf = this ;
	if (i == 0) { rootPart = atom[i]; }
    }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  wire up via connections\n");
   }

    // ... and then wire them semi-randomly together (as a ring plus random
    // additional connections to ensure full part reachability for traversals)

    for (from = 0; from < NumAtomicPerComp; from++) {
	for (i = 0; i < NumConnPerAtomic; i++) {
	    if (i == 0) {
	        to = (from + 1) % NumAtomicPerComp;
	    } else {
	        to = (int) (random() % NumAtomicPerComp);
	    }
	    conn = new (this) Connection (atom[from], atom[to]);
	}
    }

   if (debugMode) {
        printf("IN CompositePart::CompositePart:  local cleanup\n");
   }

    delete [NumAtomicPerComp] atom;
    nextAtomicId += NumAtomicPerComp;

    // finally insert this composite part as a child of the base
    // assemblies that use it

    // first the assemblies using the comp part as a shared component
    shared_cp[cpId].initScan();

    // get the first base assembly
    ba = shared_cp[cpId].next();
    while (ba != NULL)
    {
        // add this composite part cp the list of shared parts used
        // in the assembly
	ba->componentsShar += this;
	this->usedInShar += ba ;

        // continue with next base assembly
        ba = shared_cp[cpId].next();
    }

    // next the assemblies using the comp part as a private component
    private_cp[cpId].initScan();

    // get the first base assembly
    ba = private_cp[cpId].next();
    while (ba != NULL)
    {
	// add this composite part to the list of private parts used
        // in the assembly
	ba->componentsPriv += this;
	this->usedInPriv += ba ;

        // continue with next base assembly
        ba = private_cp[cpId].next();
    }

   if (debugMode) {
        printf("END CompositePart::CompositePart(cpId = %d)\n", cpId);
   }
}


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Destructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::~CompositePart()
{
    if (debugMode) {
        printf("Entering CompositePart::~CompositePart(id = %d)\n", id);
    }

    // delete associated atomic parts
    if (debugMode) {
      printf("    In ~CompositePart, about to delete parts\n");
    }
    for (int i = 0; i < parts.count(); i++) {
      delete (AtomicPart *) parts[i];
    }
    parts.clear() ;

    Iterator<BaseAssembly> iter(this->usedInShar) ;
    for (; iter.more(); iter.next()) {
      int pos = iter.get()->componentsShar.member(this) ;
      iter.get()->componentsShar.remove(pos) ;
    }

    Iterator<BaseAssembly> iter2(this->usedInPriv) ;
    for (; iter2.more(); iter2.next()) {
      int pos = iter2.get()->componentsPriv.member(this) ;
      iter2.get()->componentsPriv.remove(pos) ;
    }

    // delete associated documentation
    if (debugMode) {
      printf("    In ~CompositePart, about to delete documentation\n");
    }
    delete (Document*) documentation;

    // remove part from CompositePart extent
    if (debugMode) {
      printf("    In ~CompositePart, about to remove part from CompositePart::extent\n");
    }
    CompositePart::extent->remove(this) ;

    if (debugMode) {
        printf("Leaving CompositePart::~CompositePart(id = %d)\n", id);
    }
}

